(load "env.scm")
(load "parser.scm")

(define-struct expression ())
(define-struct (lit-exp   expression) (value))
(define-struct (var-exp   expression) (name))
(define-struct (if-exp    expression) (test then else))
(define-struct (let-exp   expression) (params args body))
(define-struct (fun-exp   expression) (params body))
(define-struct (app-exp   expression) (fun args))
(define-struct (prim-exp  expression) (prim args))

;; aspect deployment
(define-struct (arnd-exp  expression) (pc adv body))
(define-struct (farnd-exp expression) (pc adv body))

;; closure values
(define-struct value ())
(define-struct (closure value) (params body env aspects))

;; aspects
(define-struct dasp (scope pc adv))
(define-struct jp   (closure args parent))


;; evaluate expression exp in lexical environment E and aspect environment A, with current join point jp
(define (eval exp E A jp)
  (cond ((lit-exp? exp) (lit-exp-value exp))
        
        ((prim-exp? exp)
         (let ((args (eval-args (prim-exp-args exp) E A jp)))
           (apply (prim-exp-prim exp) args)))
        
        ((if-exp? exp)
         (if (eval (if-exp-test exp) E A jp)
             (eval (if-exp-then exp) E A jp)
             (eval (if-exp-else exp) E A jp)))
        
        ((let-exp? exp)
         (let ((args   (eval-args (let-exp-args exp) E A jp)))
           (eval-body (let-exp-body exp) (extend-env (let-exp-params exp) args E) A jp)))
     
        ((var-exp? exp)   (env-get (var-exp-name exp) E))
        
        ((arnd-exp? exp)  (let ((dasp (make-dasp 'static 
                                                 (eval (arnd-exp-pc exp) E A jp) 
                                                 (eval (arnd-exp-adv exp) E A jp))))
                            (eval-body (arnd-exp-body exp)  E (cons dasp A) jp)))
        
        ((farnd-exp? exp) (let ((dasp (make-dasp 'dynamic 
                                                 (eval (farnd-exp-pc exp) E A jp) 
                                                 (eval (farnd-exp-adv exp) E A jp))))
                            (eval-body (farnd-exp-body exp) E (cons dasp A) jp)))
        
        ((fun-exp? exp)   (make-closure (fun-exp-params exp) (fun-exp-body exp) E (collect-static A)))
        
        ((app-exp? exp)   (let* ((cl   (eval (app-exp-fun exp) E A jp))
                                 (args (eval-args (app-exp-args exp) E A jp))
                                 (njp  (make-jp cl args jp))
                                 (env  (extend-env (closure-params cl) args (closure-env cl)))
                                 (asps (append (collect-dynamic A) (closure-aspects cl))))
                            (weave-all A njp) 
                            (eval-body (closure-body cl) env asps njp
                                  )))))
(define (eval-args args E A jp)
    (map (lambda (x) (eval x E A jp)) args))

(define (eval-body body E A jp)
    (let loop ((exps body) (val #f))
      (if (null? exps)
          val
          (loop (cdr exps) (eval (car exps) E A jp)))))

;; aspect environments
(define (collect-static asps)  (collect-if (lambda (a) (eq? (dasp-scope a) 'static)) asps))
(define (collect-dynamic asps) (collect-if (lambda (a) (eq? (dasp-scope a) 'dynamic)) asps))


;; weaving
(define (weave-all asps jp)
  (map (lambda (x) (weave x jp)) asps))

(define (weave asp jp)
  (if (app/prim (dasp-pc asp) jp)
      (app/prim (dasp-adv asp) jp)))

;; applies the pointcut or advice function f, passing jp as argument
;; evaluation is done in an empty aspect environment
(define (app/prim f jp)
  (eval-body (closure-body f)
             (extend-env (closure-params f) (list jp)
                         (closure-env f)) '() jp))